from typing import Annotated

from depio.Task import IgnoredForEq
from depio.Task import Task


def func1(x: Annotated[int, IgnoredForEq], y: int):
    pass


def func2(x: Annotated[int, IgnoredForEq], y: int):
    pass


def test_task_eq_same_funcs():
    task1 = Task("task", func1)
    task2 = Task("task", func1)
    assert task1 == task2


def test_task_eq_diff_funcs():
    task1 = Task("task", func1)
    task2 = Task("task", func2)
    assert task1 != task2


def test_task_eq_same_args():
    task1 = Task("task", func1, [1, 2])
    task2 = Task("task", func1, [1, 2])
    assert task1 == task2


def test_task_eq_diff_args():
    task1 = Task("task", func1, [1, 2])
    task2 = Task("task", func1, [1, 3])
    assert task1 != task2


def test_task_eq_same_args_with_none():
    task1 = Task("task", func1, [None, 2])
    task2 = Task("task", func1, [None, 2])
    assert task1 == task2


def test_task_eq_same_args_with_diff_x():
    task1 = Task("task", func1, [1, 2])
    task2 = Task("task", func1, [2, 2])
    assert task1 == task2



from __future__ import annotations

import enum


class TaskStatus(enum.Enum):
    PENDING = enum.auto()
    WAITING = enum.auto()
    RUNNING = enum.auto()
    FINISHED = enum.auto()
    CANCELED = enum.auto()
    FAILED = enum.auto()
    SKIPPED = enum.auto()
    DEPFAILED = enum.auto()
    HOLD = enum.auto()
    UNKNOWN = enum.auto()


TERMINAL_STATES = [
            TaskStatus.FINISHED,
            TaskStatus.FAILED,
            TaskStatus.SKIPPED,
            TaskStatus.DEPFAILED,
            TaskStatus.CANCELED]
SUCCESSFUL_TERMINAL_STATES = [
            TaskStatus.FINISHED,
            TaskStatus.SKIPPED]
FAILED_TERMINAL_STATES = [
            TaskStatus.FAILED,
            TaskStatus.DEPFAILED,
            TaskStatus.CANCELED]



import pathlib
import os
import submitit
import unittest

from depio.Executors import SubmitItExecutor
from depio.Pipeline import Pipeline
from depio.Task import Task


def failingfunction():
    raise Exception("This function raises an exception")

def workingfunction(s: unittest.TestCase):
    s.fail("This function should never be called")

# class TestParseAnnotationForMetaclass(unittest.TestCase):
#
#     def test_failed_dependency_slurm(self):
#         SLURM = pathlib.Path("./slurm")
#         SLURM.mkdir(exist_ok=True)
#
#         # Configure the slurm jobs
#         os.environ["SBATCH_RESERVATION"] = "isec-team"
#         executor = submitit.AutoExecutor(folder=SLURM)
#         executor.update_parameters(slurm_partition="gpu", mem_gb=12, cpus_per_task=10, nodes=1, slurm_ntasks_per_node=1,
#                                    gpus_per_node=1, timeout_min=60 * 48)
#
#         pipeline = Pipeline(depioExecutor=SubmitItExecutor(internal_executor=executor))
#
#
#         t1 = Task("t1", failingfunction)
#         t2 = Task("t2", workingfunction , [self], depends_on=[t1])
#         pipeline.add_task(t2)
#         pipeline.add_task(t1)
#
#         pipeline.run()



# TASK EXCEPTIONS
class ProductNotProducedException(Exception):
    pass


class ProductNotUpdatedException(Exception):
    pass


class DependencyNotMetException(Exception):
    pass


class TaskRaisedExceptionException(Exception):
    pass


class UnknownStatusException(Exception):
    pass


# TASKHANDLER EXCEPTION
class TaskNotInQueueException(Exception):
    pass


class ProductAlreadyRegisteredException(Exception):
    pass


class DependencyNotAvailableException(Exception):
    pass



import datetime
import pathlib

def getmtime(f: pathlib.Path):
    return datetime.datetime.fromtimestamp(f.stat().st_mtime)


def getatime(f: pathlib.Path):
    return datetime.datetime.fromtimestamp(f.stat().st_atime)


def getctime(f: pathlib.Path):
    return datetime.datetime.fromtimestamp(f.stat().st_ctime)



import pytest
import pathlib

import unittest

from depio.Pipeline import Pipeline, ProductAlreadyRegisteredException, TaskNotInQueueException
from depio.Task import Task


@pytest.fixture()
def pipeline(request):
    return Pipeline(None, False, quiet=True)

def dummyfunc(self):
    pass

def test_add_task_new_task(pipeline):
    task1 = Task("task1", dummyfunc)
    pipeline.add_task(task1)
    assert task1 in pipeline.tasks

def test_add_task_duplicated_task(pipeline):
    task1 = Task("task1", dummyfunc)
    pipeline.add_task(task1)
    pipeline.add_task(task1)
    assert task1 in pipeline.tasks

def test_add_task_duplicate_producing_task(pipeline):
    producing_task = Task("producing_task", dummyfunc, produces=[pathlib.Path("test.txt")])
    producing_task2 = Task("producing_task2", dummyfunc, produces=[pathlib.Path("test.txt")])
    pipeline.add_task(producing_task)
    assert producing_task in pipeline.tasks
    pipeline.add_task(producing_task2)
    # The second task is not additionally added as it calls the same function with the same args.
    # The name is ignored.
    assert len(pipeline.tasks) == 1

def test_add_task_unregistered_dependency(pipeline):
    task1 = Task("task1", dummyfunc)
    task2 = Task("task2", dummyfunc, depends_on=[task1])
    with pytest.raises(TaskNotInQueueException):
        pipeline.add_task(task2)

def test_add_task_registered_dependency(pipeline):
    task1 = Task("task1", dummyfunc)
    pipeline.add_task(task1)
    # assert task1._queue_id == 1
    task2 = Task("task2", dummyfunc, depends_on=[task1])
    pipeline.add_task(task2)
    # assert task2._queue_id == 2
    assert task1 in pipeline.tasks
    assert task2 in pipeline.tasks



import concurrent.futures
from abc import ABC, abstractmethod
from concurrent.futures.thread import ThreadPoolExecutor
from typing import List
import submitit
from attrs import frozen
from pathlib import Path

from .Task import Task


class AbstractTaskExecutor(ABC):
    @abstractmethod
    def submit(self, task, task_dependencies: List[Task] = None):
        ...

    @abstractmethod
    def wait_for_all(self):
        ...

    @abstractmethod
    def cancel_all_jobs(self):
        ...

    @abstractmethod
    def handles_dependencies(self):
        ...


@frozen
class SequentialExecutor(AbstractTaskExecutor):
    """
    A very dumb executor, that is just executing a given task right away.
    It can be used for debugging purposes.
    """

    def submit(self, task, task_dependencies: List[Task] = None):
        print("==========================================================")
        print(f"I'm going to run task <{task.name}> now")
        print("==========================================================")
        task.run()

    def wait_for_all(self):
        pass

    def cancel_all_jobs(self):
        pass

    def handles_dependencies(self):
        return False


class ParallelExecutor(AbstractTaskExecutor):

    def __init__(self, internal_executor: concurrent.futures.Executor = None, **kwargs):
        self.internal_executor = internal_executor if internal_executor is not None else ThreadPoolExecutor()
        self.running_jobs = []
        self.running_tasks = []
        print("depio-ParallelExecutor initialized")

    def submit(self, task, task_dependencies: List[Task] = None):
        job = self.internal_executor.submit(task.run)
        self.running_jobs.append(job)
        self.running_tasks.append(task)
        return

    def wait_for_all(self):
        for job in self.running_jobs:
            job.result()

    def cancel_all_jobs(self):
        pass

    def handles_dependencies(self):
        return False


TWO_DAYS_IN_MINUTES = 60 * 48  # 48 hours in minutes
DEFAULT_PARAMS = {
    "slurm_time": TWO_DAYS_IN_MINUTES,
    "slurm_partition": "gpu",
    "slurm_mem": 32,
    "gpus_per_node": 0
}


class SubmitItExecutor(AbstractTaskExecutor):

    def __init__(self, folder: Path = None, internal_executor=None, parameters=None):

        # Overwrite with a default executor.
        if internal_executor is None:
            internal_executor = submitit.AutoExecutor(folder=folder)
            internal_executor.update_parameters(**parameters)

        self.internal_executor = internal_executor
        self.default_parameters = parameters if parameters is not None else DEFAULT_PARAMS
        self.internal_executor.update_parameters(**self.default_parameters)

        self.slurmjobs = []
        print("depio-SubmitItExecutor initialized")

    def submit(self, task, task_dependencies: List[Task] = None):
        slurm_additional_parameters = {}

        tasks_with_slurmjob = [t for t in task_dependencies if t.slurmjob is not None]
        afterok: List[str] = [f"{t.slurmjob.job_id}" for t in tasks_with_slurmjob]

        if len(afterok) > 0:
            slurm_additional_parameters["dependency"] = f"afterok:{':'.join(afterok)}"

        if task.slurm_parameters is not None:
            params = task.slurm_parameters
        else:
            params = self.default_parameters
        self.internal_executor.update_parameters(**params, slurm_additional_parameters=slurm_additional_parameters)

        slurmjob = self.internal_executor.submit(task.run)
        task.slurmjob = slurmjob
        self.slurmjobs.append(slurmjob)
        return

    def wait_for_all(self):
        for job in self.slurmjobs:
            job.result()

    def cancel_all_jobs(self):
        for job in self.slurmjobs:
            job.cancel()

    def handles_dependencies(self):
        return True


__all__ = [AbstractTaskExecutor, ParallelExecutor, SequentialExecutor, SubmitItExecutor]



# Python
import pytest
from depio.Task import Task, _status_colors, UnknownStatusException


def test_statuscolor():
    test_task = Task("test_task", lambda: print("Hello World"))  # Creating a dummy Task object

    for key in _status_colors.keys():
        assert test_task.statuscolor(key) == _status_colors[key], f"Expected task status color for key '{key}' to be '{_status_colors[key]}'"

    with pytest.raises(UnknownStatusException):
        test_task.statuscolor('unrecognized_status')  # Will throw an exception as the status is not recognized.



# test_Task__parse_annotation_for_metaclass.py

import unittest
from typing import Annotated

from depio.Task import _parse_annotation_for_metaclass


class A: pass  # Dummy metaclass for testing


class TestParseAnnotationForMetaclass(unittest.TestCase):

    def test_parse_annotation_for_metaclass(self):
        """Test _parse_annotation_for_metaclass"""

        def dummy_function(inputa :Annotated[str, A]) -> None: pass

        expected_result = ['inputa']  # 'return' is the annotation name for function return type
        parsed_annotations = _parse_annotation_for_metaclass(dummy_function, A)
        self.assertEqual(parsed_annotations, expected_result)

    def test_parse_annotation_for_metaclass_return(self):
        """Test _parse_annotation_for_metaclass"""

        def dummy_function() -> Annotated[str, A]: pass

        expected_result = ['return']  # 'return' is the annotation name for function return type
        parsed_annotations = _parse_annotation_for_metaclass(dummy_function, A)
        self.assertEqual(parsed_annotations, expected_result)

    def test_parse_annotation_for_metaclass_no_args(self):
        """Test _parse_annotation_for_metaclass"""

        def dummy_function() -> None: pass

        parsed_annotations = _parse_annotation_for_metaclass(dummy_function, A)
        self.assertEqual(parsed_annotations, [])

    def test_parse_annotation_for_metaclass_no_metaclass(self):
        """Test _parse_annotation_for_metaclass when no matching metaclass in annotation"""

        # Change metaclass in function annotation
        def dummy_function() -> Annotated[str, list]: pass

        parsed_annotations = _parse_annotation_for_metaclass(dummy_function, A)
        self.assertEqual(parsed_annotations, [])

    def test_parse_annotation_for_metaclass_no_annotations(self):
        """Test _parse_annotation_for_metaclass when no function annotations"""

        def dummy_function(a, b): pass  # No type annotations

        parsed_annotations = _parse_annotation_for_metaclass(dummy_function, A)
        self.assertEqual(parsed_annotations, [])


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from pathlib import Path
import typing
import time
from io import StringIO
from os.path import getmtime
from typing import List, Dict, Callable, get_origin, Annotated, get_args, Union
import sys

from .BuildMode import BuildMode
from .TaskStatus import TaskStatus, TERMINAL_STATES, SUCCESSFUL_TERMINAL_STATES, FAILED_TERMINAL_STATES
from .stdio_helpers import redirect, stop_redirect
from .exceptions import ProductNotProducedException, TaskRaisedExceptionException, UnknownStatusException, \
    ProductNotUpdatedException, \
    DependencyNotMetException


class Product():
    pass


class Dependency():
    pass


class IgnoredForEq():
    pass


_status_colors = {
    TaskStatus.WAITING: 'blue',
    TaskStatus.DEPFAILED: 'red',
    TaskStatus.PENDING: 'blue',
    TaskStatus.RUNNING: 'yellow',
    TaskStatus.FINISHED: 'green',
    TaskStatus.SKIPPED: 'green',
    TaskStatus.HOLD: 'white',
    TaskStatus.FAILED: 'red',
    TaskStatus.CANCELED: 'white',
    TaskStatus.UNKNOWN: 'white'
}

_status_texts = {
    TaskStatus.PENDING: 'pending',
    TaskStatus.RUNNING: 'running',
    TaskStatus.FINISHED: 'finished',
    TaskStatus.SKIPPED: 'skipped',
    TaskStatus.HOLD: 'hold',
    TaskStatus.FAILED: 'failed',
    TaskStatus.CANCELED: 'cancelled',
    TaskStatus.UNKNOWN: 'unknown',
    TaskStatus.WAITING: 'waiting',
    TaskStatus.DEPFAILED: 'dep. failed'
}


def python_version_is_greater_or_equal_to_3_10():
    return sys.version_info.major > 3 and sys.version_info.minor >= 10


# from https://stackoverflow.com/questions/218616/how-to-get-method-parameter-names
def _get_args_dict(fn, args, kwargs) -> Dict[str, typing.Any]:
    args_names = fn.__code__.co_varnames[:fn.__code__.co_argcount]
    return {**dict(zip(args_names, args)), **kwargs}


def _parse_annotation_for_metaclass(func, metaclass) -> List[str]:
    if python_version_is_greater_or_equal_to_3_10():
        # For python 3.10 and newer
        # annotations = inspect.get_annotations(func)

        # According to https://docs.python.org/3/howto/annotations.html this is best practice now.
        annotations = getattr(func, '__annotations__', None)
    else:
        # For python 3.9 and older
        if isinstance(func, type):
            annotations = func.__dict__.get('__annotations__', None)
        else:
            annotations = getattr(func, '__annotations__', None)

    results: List[str] = []

    for name, annotation in annotations.items():
        if get_origin(annotation) is Annotated:
            args = get_args(annotation)
            if len(args) <= 1:
                continue

            metadata = args[1:]
            if any(meta is metaclass for meta in metadata):
                results.append(name)

    return results


def _get_not_updated_products(product_timestamps_after_running: typing.Dict,
                              product_timestamps_before_running: typing.Dict) -> typing.List[str]:
    # Calculate the not updated products
    not_updated_products = []
    for product, before_timestamp in product_timestamps_before_running.items():
        after_timestamp = product_timestamps_after_running.get(product)
        if before_timestamp == after_timestamp:
            not_updated_products.append(product)

    return not_updated_products


class Task:
    def __init__(self, name: str, func: Callable, func_args: List = None, func_kwargs: List = None,
                 produces: List[Path] = None, depends_on: List[Union[Path, Task]] = None,
                 buildmode: BuildMode = BuildMode.IF_MISSING,
                 slurm_parameters: Dict = None,
                 arg_resolver: Callable = None,
                 description: str = None):

        self.end_time = None
        self.start_time = None
        self.description = description or ""
        produces: List[Path] = produces or []
        depends_on: List[Union[Path, Task]] = depends_on or []

        self._status: TaskStatus = TaskStatus.WAITING
        self.name: str = name
        self._queue_id: int | None = None
        self.slurmjob = None
        self.func: Callable = func
        self.func_args: List = func_args or []
        self.func_kwargs: Dict = func_kwargs or {}
        self.buildmode: BuildMode = buildmode
        self.slurm_parameters: Dict = slurm_parameters or {}

        self.stdout: StringIO = StringIO()
        self.stderr: StringIO = StringIO()
        self.slurmjob = None
        self._slurmid = None
        self._slurmstate: str = ""

        # Parse dependencies and products from the annotations and merge with args
        products_args: List[str] = _parse_annotation_for_metaclass(func, Product)
        dependencies_args: List[str] = _parse_annotation_for_metaclass(func, Dependency)
        ignored_for_eq_args: List[str] = _parse_annotation_for_metaclass(func, IgnoredForEq)

        # Allow the task to specify an argument resolver. This can be used to load default values dynamically.
        # And in particular, before the DAG is constructed.
        if arg_resolver is not None:
            self.func_args, self.func_kwargs = arg_resolver(self.func, self.func_args, self.func_kwargs)

        args_dict: Dict[str, typing.Any] = _get_args_dict(func, self.func_args, self.func_kwargs)
        self.cleaned_args: Dict[str, typing.Any] = {k: v for k, v in args_dict.items() if k not in ignored_for_eq_args}

        self.products: List[Path] = \
            ([args_dict[argname] for argname in products_args if argname in args_dict and args_dict[argname] is not None] + produces)
        self.dependencies: List[Union[Task, Path]] = \
            ([args_dict[argname] for argname in dependencies_args if argname in args_dict and args_dict[argname] is not None] + depends_on)

        # Gets filled by Pipeline
        self.path_dependencies = None
        self.task_dependencies = None

        self.dependent_tasks = []

    def is_ready_for_execution(self) -> bool:
        if not self.should_run():
            self.set_to_skipped()
            return False

        if not self.all_path_dependencies_exist() and not self.is_in_terminal_state:
            self.set_to_depfailed()
            return False

        if self.is_in_terminal_state:
            return False

        return self.all_path_dependencies_exist() and self.all_task_dependencies_terminated_successfully()


    def all_path_dependencies_exist(self) -> bool:
        return all(p_dep.exists() for p_dep in self.path_dependencies)

    def all_task_dependencies_terminated_successfully(self) -> bool:
        return all(t_dep.is_in_successful_terminal_state for t_dep in self.task_dependencies)

    def add_dependent_task(self, task):
        self.dependent_tasks.append(task)

    def __str__(self):
        return f"Task:{self.name}"

    def should_run(self) -> bool:
        missing_products: List[Path] = [p for p in self.products if not p.exists()]

        if self.buildmode == BuildMode.ALWAYS:
            return True
        elif self.buildmode == BuildMode.IF_MISSING:
            return len(missing_products) > 0
        elif self.buildmode == BuildMode.IF_NEW:
            return any(t.should_run() for t in self.task_dependencies) or len(missing_products) > 0
        elif self.buildmode == BuildMode.NEVER:
            return False
        else:
            raise Exception(f"Unkown buildmode: {self.buildmode}")

    def _check_path_dependencies(self):
        not_existing_path_dependencies: List[str] = \
            [str(dependency) for dependency in self.path_dependencies if not dependency.exists()]

        if len(not_existing_path_dependencies) > 0:
            self._status = TaskStatus.FAILED
            raise DependencyNotMetException(
                f"Task {self.name}: Dependency/ies {not_existing_path_dependencies} not met.")

    def _check_existence_of_products(self):
        not_existing_products: List[str] = [str(product) for product in self.products if not product.exists()]
        if len(not_existing_products) > 0:
            self._status = TaskStatus.FAILED
            raise ProductNotProducedException(f"Task {self.name}: Product/s {not_existing_products} not produced.")

    def _get_timestamp_of_products(self) -> Dict[str, float]:
        return {str(product): getmtime(product) for product in self.products if product.exists()}

    def get_duration(self) -> int:
        if self.start_time is None:
            return 0 # secs
        if self.end_time is None:
            return int(time.time() - self.start_time)
        return int(self.end_time - self.start_time)

    def run(self):
        self.start_time = time.time()
        redirect(self.stdout)

        # Check if all path dependencies are met
        self._check_path_dependencies()

        # Store the last-modification timestamp of the already existing products.
        product_timestamps_before_running: Dict[str, float] = self._get_timestamp_of_products()

        # Call the actual function
        self._status = TaskStatus.RUNNING

        try:
            self.func(*self.func_args, **self.func_kwargs)
        except Exception as e:
            self.set_to_failed()
            raise TaskRaisedExceptionException(e)
        finally:
            stop_redirect()

        # Check if any product does not exist.
        self._check_existence_of_products()

        # Check if any product has not been updated.
        product_timestamps_after_running: Dict[str, float] = self._get_timestamp_of_products()
        not_updated_products = _get_not_updated_products(product_timestamps_after_running,
                                                         product_timestamps_before_running)
        if len(not_updated_products) > 0:
            self._status = TaskStatus.FAILED
            raise ProductNotUpdatedException(f"Task {self.name}: Product/s {not_updated_products} not updated.")

        self._status = TaskStatus.FINISHED
        self.end_time = time.time()

    def barerun(self):
        self.func(*self.func_args, **self.func_kwargs)

    def _set_status_by_slurmstate(self, slurmstate):

        if slurmstate in ['RUNNING', 'CONFIGURING', 'COMPLETING', 'STAGE_OUT']:
            _status = TaskStatus.RUNNING
        elif slurmstate in ['FAILED', 'BOOT_FAIL', 'DEADLINE', 'NODE_FAIL', 'OUT_OF_MEMORY',
                            'PREEMPTED', 'SPECIAL_EXIT', 'STOPPED', 'SUSPENDED', 'TIMEOUT']:
            _status = TaskStatus.FAILED
            self.set_to_failed()
        elif slurmstate in ['READY', 'PENDING', 'REQUEUE_FED', 'REQUEUED']:
            _status = TaskStatus.PENDING
        elif slurmstate == 'CANCELLED':
            _status = TaskStatus.CANCELED
        elif 'CANCELLED' in slurmstate: # Slurm set 'CANCELLED by <number>' sometimes...
            _status = TaskStatus.CANCELED
        elif slurmstate in ['COMPLETED']:
            _status = TaskStatus.FINISHED
        elif slurmstate in ['RESV_DEL_HOLD', 'REQUEUE_HOLD', 'RESIZING', 'REVOKED', 'SIGNALING']:
            _status = TaskStatus.HOLD
        elif slurmstate in ['UNKNOWN']:
            _status = TaskStatus.UNKNOWN
        else:
            #print(f"Unknown slurmjob status! -> {slurmstate} ")
            _status = TaskStatus.UNKNOWN

        self._status = _status
        return _status

    def _update_by_slurmjob(self):
        assert self.slurmjob is not None

        self.slurmjob.watcher.update()

        self._slurmstate = self.slurmjob.state
        self._set_status_by_slurmstate(self._slurmstate)

        self._slurmid = f"{int(self.slurmjob.job_id):d}-{int(self.slurmjob.task_id):d}"

    @property
    def slurmjob_status(self):
        if self.slurmjob is None: return ""

        if self._slurmstate is None: self._update_by_slurmjob()
        return self._slurmstate

    def statuscolor(self, s: TaskStatus = None) -> str:
        if s is None: s = self._status
        if s in _status_colors:
            return _status_colors[s]
        else:
            raise UnknownStatusException(f"Status {s} is unknown.")

    def statustext(self, s: TaskStatus = None) -> str:
        if s is None: s = self._status
        if s in _status_texts:
            return _status_texts[s]

        raise UnknownStatusException(f"Status {s} is unknown.")

    def statustext_long(self, s: TaskStatus = None) -> str:
        if s is None: s = self._status
        if s in _status_texts:
            return _status_texts[s]

        status_messages = {
            TaskStatus.WAITING: lambda: 'waiting' + (
                f" for {[d._queue_id for d in self.task_dependencies if not d.is_in_terminal_state]}" if len(
                    [d for d in self.task_dependencies if not d.is_in_terminal_state]) > 1 else ""),
            TaskStatus.DEPFAILED: lambda: 'dep. failed' + (
                f" at {[d._queue_id for d in self.task_dependencies if d.is_in_failed_terminal_state]}" if len(
                    [d for d in self.task_dependencies if d.is_in_failed_terminal_state]) > 1 else "")
        }
        try:
            return status_messages[s]()
        except KeyError:
            raise UnknownStatusException(f"Status {s} is unknown.")

    @property
    def status(self):
        if self.slurmjob is None:
            s = self._status
            slurmstate = ""
        else:
            if self._slurmstate is None: self._update_by_slurmjob()
            slurmstate = self._slurmstate
            s = self._set_status_by_slurmstate(slurmstate)
        return s, self.statustext(s), self.statuscolor(s), slurmstate

    @property
    def is_in_terminal_state(self) -> bool:
        return self._status in TERMINAL_STATES

    @property
    def is_in_successful_terminal_state(self) -> bool:
        return self._status in SUCCESSFUL_TERMINAL_STATES

    @property
    def is_in_failed_terminal_state(self) -> bool:
        return self._status in FAILED_TERMINAL_STATES

    def set_dependent_task_to_depfailed(self):
        for task in self.dependent_tasks:
            task.set_to_depfailed()

    def set_to_failed(self):
        self._status = TaskStatus.FAILED
        if self.slurmjob is not None:
            self.slurmjob.cancel()
        self.set_dependent_task_to_depfailed()

    def set_to_depfailed(self) -> None:
        self._status = TaskStatus.DEPFAILED
        if self.slurmjob is not None:
            self.slurmjob.cancel()
        self.set_dependent_task_to_depfailed()


    def set_to_skipped(self) -> None:
        self._status = TaskStatus.SKIPPED

    @property
    def id(self) -> str:
        if self._queue_id:
            return f"{self._queue_id: 4d}"
        else:
            return "None"

    @property
    def slurmid(self) -> str:
        if self.slurmjob is None:
            return ""

        self._update_by_slurmjob()
        return f"{self._slurmid}"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.func != other.func:
                return False

            for k,v1 in self.cleaned_args.items():
                if k not in other.cleaned_args:
                    return False
                if v1 is None and other.cleaned_args[k] is not None:
                    return False
                if v1 is not None and other.cleaned_args[k] is None:
                    return False
                if v1 is None and other.cleaned_args[k] is None:
                    continue # None and None should be considered equal!
                if v1 != other.cleaned_args[k]:
                    return False
                # => v1 = other.cleaned_args[k] or both are None

            # And the other way around....
            for k,v1 in other.cleaned_args.items():
                if k not in self.cleaned_args:
                    return False
                if v1 is None and self.cleaned_args[k] is not None:
                    return False
                if v1 is not None and self.cleaned_args[k] is None:
                    return False
                if v1 is None and self.cleaned_args[k] is None:
                    continue # None and None should be considered equal!
                if v1 != self.cleaned_args[k]:
                    return False
                # => v1 = self.cleaned_args[k] or both are None

            return True


        else:
            return False

    def get_stderr(self):
        if self.slurmjob is None:
            return self.stderr.getvalue()
        else:
            return self.slurmjob.stderr()

    def get_stdout(self):
        if self.slurmjob is None:
            return self.stdout.getvalue()
        else:
            return self.slurmjob.stdout()


__all__ = [Task, Product, Dependency, _get_not_updated_products]



from .BuildMode import BuildMode
from .Pipeline import Pipeline
from .Task import Task


def task(name: str | None, pipeline: Pipeline | None = None, buildmode: BuildMode = BuildMode.IF_MISSING, *dec_args,
         **dec_kwargs):
    def wrapper(func):
        def decorator(*func_args, **func_kwargs):
            # Create and add the task
            t = Task(name, func=func, buildmode=buildmode, func_args=func_args, func_kwargs=func_kwargs)

            # Not call the function!

            if pipeline:
                pipeline.add_task(t)

            return t

        return decorator

    return wrapper


__all__ = [task]



import pathlib
from distutils.command.build_scripts import first_line_re
from logging import setLoggerClass
from typing import Set, Dict, List
from pathlib import Path
import time
import sys


from termcolor import colored
from tabulate import tabulate

from .stdio_helpers import enable_proxy
from .Task import Task
from .TaskStatus import TaskStatus
from .Executors import AbstractTaskExecutor
from .exceptions import ProductAlreadyRegisteredException, TaskNotInQueueException, DependencyNotAvailableException


class Pipeline:
    def __init__(self, depioExecutor: AbstractTaskExecutor, name: str = "NONAME",
                 clear_screen: bool = True,
                 hide_successful_terminated_tasks: bool = False,
                 submit_only_if_runnable: bool = False,
                 quiet: bool = False,
                 refreshrate: float = 1.0):

        # Flags
        self.CLEAR_SCREEN: bool = clear_screen
        self.QUIET: bool = quiet
        self.REFRESHRATE: float = refreshrate
        self.HIDE_SUCCESSFUL_TERMINATED_TASKS: bool = hide_successful_terminated_tasks
        self.SUBMIT_ONLY_IF_RUNNABLE :bool = submit_only_if_runnable

        self.name: str = name
        self.handled_tasks: List[Task] = None
        self.tasks: List[Task] = []
        self.depioExecutor: AbstractTaskExecutor = depioExecutor
        self.registered_products: Set[Path] = set()
        if not self.QUIET: print("Pipeline initialized")

    def add_tasks(self, tasks: List[Task]) -> None:
        for task in tasks:
            self.add_task(task)

    def add_task(self, task: Task) -> None:

        # Check if the exact task is already registered
        for registered_task in self.tasks:
            if task == registered_task:
                return registered_task


        # Check is a product is already registered
        products_already_registered: List[str] = [str(p) for p in task.products if
                                                  str(p) in set(map(str, self.registered_products))]
        if len(products_already_registered) > 0:
            print(task.cleaned_args)
            for p in products_already_registered:
                t = [t for t in self.tasks if str(p) in set(map(str, t.products))][0]
                print(f"Product {p} is already registered by task {t.name}. Now again registered by task {task.name}.")
            raise ProductAlreadyRegisteredException(
                f"The product/s {products_already_registered} is/are already registered. "
                f"Each output can only be registered from one task.")


        # Check if the task dependencies are registered already
        missing_tasks: List[Task] = [t for t in task.dependencies if isinstance(t, Task) and t not in self.tasks]
        if len(missing_tasks) > 0:
            raise TaskNotInQueueException(f"Add the tasks into the queue in the correct order. "
                                          f"The following task/s is/are missing: {missing_tasks}.")

        # Register products
        self.registered_products.update(task.products)

        # Register task
        self.tasks.append(task)
        task._queue_id = len(self.tasks)  # TODO Fix this!
        return task

    def _solve_order(self) -> None:
        # Generate a task to product mapping.
        product_to_task: Dict[Path, Task] = {product: task for task in self.tasks for product in task.products}

        # Add the dependencies to the tasks
        for task in self.tasks:
            # First spit of into tasks and paths
            task_deps = [d for d in task.dependencies if isinstance(d, Task)]
            path_deps = [d for d in task.dependencies if isinstance(d, Path)]

            # Verify that each dependency is available and add if yes.
            unavailable_dependencies = [d for d in path_deps if d not in product_to_task and not d.exists()]
            if len(unavailable_dependencies) > 0:
                raise DependencyNotAvailableException(f"Dependency/ies '{unavailable_dependencies}' "
                                                      f"do/es not exist and can not be produced.")

            # Add the tasks that produce path_deps and remove such deps from the path_deps
            task.task_dependencies = []
            for t in ([product_to_task[d] for d in path_deps if d in product_to_task] + task_deps):
                if not t in task.task_dependencies: task.task_dependencies.append(t)

            task.path_dependencies = \
                [d for d in path_deps if d not in product_to_task]

        # Adding the backlinks
        for task in self.tasks:
            for t_dep in task.task_dependencies:
                t_dep.add_dependent_task(task)

    def run(self) -> None:
        enable_proxy()
        self._solve_order()
        self.handled_tasks = []
        while True:
            try:
                # Submit new runnable jobs.
                for task in self.tasks:
                    if task in self.handled_tasks: continue
                    if task.is_ready_for_execution() or self.depioExecutor.handles_dependencies():
                        if task.should_run():
                            if not self.SUBMIT_ONLY_IF_RUNNABLE:
                                self.depioExecutor.submit(task, task.task_dependencies)
                                self.handled_tasks.append(task)
                            elif task.is_ready_for_execution():
                                if len([t for t  in self.handled_tasks if t.status[0] in [TaskStatus.PENDING,TaskStatus.UNKNOWN]]) > 20:
                                    pass # Only have n jobs in pending. # TODO Make the 20 a variable n!
                                elif len([t for t  in self.handled_tasks if not t.is_in_terminal_state]) > 45:
                                    pass # Only have at most 45 jobs in non-terminal states. (Do not overload the BWUniCluster)
                                else:
                                    self.depioExecutor.submit(task, task.task_dependencies)
                                    self.handled_tasks.append(task)
                            else:
                                pass


                # Check the status of all tasks
                if not self.QUIET: self._print_tasks()
                if all(task.is_in_terminal_state for task in self.tasks):
                    if any(task.is_in_failed_terminal_state for task in self.tasks):
                        self.exit_with_failed_tasks()
                    else:
                        self.exit_successful()

                time.sleep(self.REFRESHRATE)

            except KeyboardInterrupt:
                print("Stopping execution because of a keyboard interrupt!")
                self.exit_with_failed_tasks()

    def _get_text_for_task(self, task):
        status = task.status

        formatted_status = colored(f"{status[1].upper():<{len('DEP. FAILED')}s}", status[2])
        formatted_slurmstatus = colored(f"{status[3]:<{len('OUT_OF_MEMORY')}s}", status[2])
        return [
            task.is_in_successful_terminal_state,
            task.id,
            task.name,
            task.slurmid,
            formatted_slurmstatus,
            formatted_status,
            #task.get_duration(),
            [t._queue_id for t in task.task_dependencies],
            #[str(d) for d in task.dependencies if isinstance(d, Path)],
            #[str(p) for p in task.products]
        ]

    def _clear_screen(self):
        if self.CLEAR_SCREEN: sys.stdout.write("\033[2J\033[H")

    def _print_tasks(self):
        self._clear_screen()
        headers = ["ID", "Name", "Slurm ID", "Slurm Status", "Status", "Task Deps."]

        tasks_data = [self._get_text_for_task(task) for task in self.tasks]

        # Generate and print the histogram for tasks_data[5]
        histogram = {}
        for task_data in tasks_data:
            s = task_data[5]
            histogram[s] = histogram.get(s, 0) + 1

        if self.HIDE_SUCCESSFUL_TERMINATED_TASKS:
            tasks_data = [td[1:] for td in tasks_data if not td[0] == True]
        else:
            tasks_data = [td[1:] for td in tasks_data]

        table_str = tabulate(tasks_data, headers=headers, tablefmt="plain")
        print()
        print("Tasks:")
        print(table_str)

        print("\nSummary:")
        for string, count in histogram.items():
            print(f"{string}: {count:4d} tasks")


    def exit_with_failed_tasks(self) -> None:
        print()

        # Print the overview with the updated status once more.
        for task in self.tasks:
            task.is_ready_for_execution()
        if not self.QUIET: self._print_tasks()


        failed_tasks = [
            [task.id, task.name, task.slurmid, task.status[1]]
            for task in self.tasks if task.status[0] == TaskStatus.FAILED
        ]

        if failed_tasks:
            headers = ["Task ID", "Name", "Slurm ID", "Status"]
            print("---> Summary of Failed Tasks:")
            print()

            for task in self.tasks:
                if task.status[0] == TaskStatus.FAILED:
                    print(f"Details for Task ID: {task.id} - Name: {task.name}")
                    print(f"STDOUT")
                    print(task.get_stdout())
                    print(f"")
                    print(f"STDERR")
                    print(task.get_stderr())

        print("Canceling running jobs...")
        self.depioExecutor.cancel_all_jobs()

        print("Exit.")
        exit(1)

    def exit_successful(self) -> None:
        # Print the overview with the updated status once more.
        for task in self.tasks:
            task.is_ready_for_execution()
        if not self.QUIET: self._print_tasks()

        print("All jobs done! Exit.")
        exit(0)


__all__ = [Pipeline]



from depio.Task import Task


def func1(a: int, b: int, c: int):
    pass


def func2(a: int, b: int, c: int):
    pass


def test_task_eq_same_funcs():
    task1 = Task("task", func1)
    task2 = Task("task", func1)
    assert task1 == task2


def test_task_eq_diff_funcs():
    task1 = Task("task", func1)
    task2 = Task("task", func2)
    assert task1 != task2


def test_task_eq_same_args():
    task1 = Task("task", func1, [1, 2, 3])
    task2 = Task("task", func1, [1, 2, 3])
    assert task1 == task2


def test_task_eq_diff_args():
    task1 = Task("task", func1, [1, 2, 3])
    task2 = Task("task", func1, [1, 2, 4])
    assert task1 != task2


def test_task_eq_same_args_with_none():
    task1 = Task("task", func1, [None, 2, 3])
    task2 = Task("task", func1, [None, 2, 3])
    assert task1 == task2


def test_task_eq_diff_args_with_none():
    task1 = Task("task", func1, [None, 2, 3])
    task2 = Task("task", func1, [None, 2, 4])
    assert task1 != task2


def test_task_eq_same_kwargs():
    task1 = Task("task", func1, None, {'a': 1, 'b': 2})
    task2 = Task("task", func1, None, {'a': 1, 'b': 2})
    assert task1 == task2


def test_task_eq_diff_kwargs():
    task1 = Task("task", func1, None, {'a': 1, 'b': 2})
    task2 = Task("task", func1, None, {'a': 1, 'b': 3})
    assert task1 != task2


def test_task_eq_same_name():
    task1 = Task("task1", func1)
    task2 = Task("task1", func1)
    assert task1 == task2





# from https://gitlab.com/yquemener/stdout-redirects with some adaptions.
# copied from https://stackoverflow.com/a/43667367/1193986
#
# (c) umichscoots 2017
# License unsepcified. Assumed to be CC-by-sa as is StackOverflow's policy
#
# The class LocalProxy is taken from the werkzeug project
# https://raw.githubusercontent.com/pallets/werkzeug/ef545f0d0bf28cbad02066b4cb7471bea50a93ee/src/werkzeug/local.py
# It is licensed under the BSD-3-Clause License
#
# I guess that means the result is CC-by-SA


import threading
import sys
from io import StringIO
import copy

from typing import Any
from typing import Optional
from typing import Union

# Save all of the objects for use later.
orig___stdout__ = sys.__stdout__
orig___stderr__ = sys.__stderr__
orig_stdout = sys.stdout
orig_stderr = sys.stderr
thread_proxies = {}


class LocalProxy:
    """Acts as a proxy for a werkzeug local.  Forwards all operations to
    a proxied object.  The only operations not supported for forwarding
    are right handed operands and any kind of assignment.
    Example usage::
        from werkzeug.local import Local
        l = Local()
        # these are proxies
        request = l('request')
        user = l('user')
        from werkzeug.local import LocalStack
        _response_local = LocalStack()
        # this is a proxy
        response = _response_local()
    Whenever something is bound to l.user / l.request the proxy objects
    will forward all operations.  If no object is bound a :exc:`RuntimeError`
    will be raised.
    To create proxies to :class:`Local` or :class:`LocalStack` objects,
    call the object as shown above.  If you want to have a proxy to an
    object looked up by a function, you can (as of Werkzeug 0.6.1) pass
    a function to the :class:`LocalProxy` constructor::
        session = LocalProxy(lambda: get_current_request().session)
    .. versionchanged:: 0.6.1
       The class can be instantiated with a callable as well now.
    """

    __slots__ = ("__local", "__dict__", "__name__", "__wrapped__")

    def __init__(
            self, local: Union[Any, "LocalProxy"], name: Optional[str] = None,
    ) -> None:
        object.__setattr__(self, "_LocalProxy__local", local)
        object.__setattr__(self, "__name__", name)
        if callable(local) and not hasattr(local, "__release_local__"):
            # "local" is a callable that is not an instance of Local or
            # LocalManager: mark it as a wrapped function.
            object.__setattr__(self, "__wrapped__", local)

    def _get_current_object(self, ) -> object:
        """Return the current object.  This is useful if you want the real
        object behind the proxy at a time for performance reasons or because
        you want to pass the object into a different context.
        """
        if not hasattr(self.__local, "__release_local__"):
            return self.__local()
        try:
            return getattr(self.__local, self.__name__)
        except AttributeError:
            raise RuntimeError(f"no object bound to {self.__name__}")

    @property
    def __dict__(self):
        try:
            return self._get_current_object().__dict__
        except RuntimeError:
            raise AttributeError("__dict__")

    def __repr__(self) -> str:
        try:
            obj = self._get_current_object()
        except RuntimeError:
            return f"<{type(self).__name__} unbound>"
        return repr(obj)

    def __bool__(self) -> bool:
        try:
            return bool(self._get_current_object())
        except RuntimeError:
            return False

    def __dir__(self):
        try:
            return dir(self._get_current_object())
        except RuntimeError:
            return []

    def __getattr__(self, name: str) -> Any:
        if name == "__members__":
            return dir(self._get_current_object())
        return getattr(self._get_current_object(), name)

    def __setitem__(self, key: Any, value: Any) -> None:
        self._get_current_object()[key] = value  # type: ignore

    def __delitem__(self, key):
        del self._get_current_object()[key]

    __setattr__ = lambda x, n, v: setattr(x._get_current_object(), n, v)  # type: ignore
    __delattr__ = lambda x, n: delattr(x._get_current_object(), n)  # type: ignore
    __str__ = lambda x: str(x._get_current_object())  # type: ignore
    __lt__ = lambda x, o: x._get_current_object() < o
    __le__ = lambda x, o: x._get_current_object() <= o
    __eq__ = lambda x, o: x._get_current_object() == o  # type: ignore
    __ne__ = lambda x, o: x._get_current_object() != o  # type: ignore
    __gt__ = lambda x, o: x._get_current_object() > o
    __ge__ = lambda x, o: x._get_current_object() >= o
    __hash__ = lambda x: hash(x._get_current_object())  # type: ignore
    __call__ = lambda x, *a, **kw: x._get_current_object()(*a, **kw)
    __len__ = lambda x: len(x._get_current_object())
    __getitem__ = lambda x, i: x._get_current_object()[i]
    __iter__ = lambda x: iter(x._get_current_object())
    __contains__ = lambda x, i: i in x._get_current_object()
    __add__ = lambda x, o: x._get_current_object() + o
    __sub__ = lambda x, o: x._get_current_object() - o
    __mul__ = lambda x, o: x._get_current_object() * o
    __floordiv__ = lambda x, o: x._get_current_object() // o
    __mod__ = lambda x, o: x._get_current_object() % o
    __divmod__ = lambda x, o: x._get_current_object().__divmod__(o)
    __pow__ = lambda x, o: x._get_current_object() ** o
    __lshift__ = lambda x, o: x._get_current_object() << o
    __rshift__ = lambda x, o: x._get_current_object() >> o
    __and__ = lambda x, o: x._get_current_object() & o
    __xor__ = lambda x, o: x._get_current_object() ^ o
    __or__ = lambda x, o: x._get_current_object() | o
    __div__ = lambda x, o: x._get_current_object().__div__(o)
    __truediv__ = lambda x, o: x._get_current_object().__truediv__(o)
    __neg__ = lambda x: -(x._get_current_object())
    __pos__ = lambda x: +(x._get_current_object())
    __abs__ = lambda x: abs(x._get_current_object())
    __invert__ = lambda x: ~(x._get_current_object())
    __complex__ = lambda x: complex(x._get_current_object())
    __int__ = lambda x: int(x._get_current_object())
    __long__ = lambda x: long(x._get_current_object())  # type: ignore # noqa
    __float__ = lambda x: float(x._get_current_object())
    __oct__ = lambda x: oct(x._get_current_object())
    __hex__ = lambda x: hex(x._get_current_object())
    __index__ = lambda x: x._get_current_object().__index__()
    __coerce__ = lambda x, o: x._get_current_object().__coerce__(x, o)
    __enter__ = lambda x: x._get_current_object().__enter__()
    __exit__ = lambda x, *a, **kw: x._get_current_object().__exit__(*a, **kw)
    __radd__ = lambda x, o: o + x._get_current_object()
    __rsub__ = lambda x, o: o - x._get_current_object()
    __rmul__ = lambda x, o: o * x._get_current_object()
    __rdiv__ = lambda x, o: o / x._get_current_object()
    __rtruediv__ = __rdiv__
    __rfloordiv__ = lambda x, o: o // x._get_current_object()
    __rmod__ = lambda x, o: o % x._get_current_object()
    __rdivmod__ = lambda x, o: x._get_current_object().__rdivmod__(o)
    __copy__ = lambda x: copy.copy(x._get_current_object())
    __deepcopy__ = lambda x, memo: copy.deepcopy(x._get_current_object(), memo)


def redirect(stringio: StringIO):
    """
    Enables the redirect for the current thread's output to a single cStringIO
    object and returns the object.

    :return: The StringIO object.
    :rtype: ``cStringIO.StringIO``
    """
    # Get the current thread's identity.
    ident = threading.current_thread().ident

    # Enable the redirect and return the cStringIO object.
    thread_proxies[ident] = stringio


def stop_redirect():
    """
    Enables the redirect for the current thread's output to a single cStringIO
    object and returns the object.

    :return: The final string value.
    :rtype: ``str``
    """
    # Get the current thread's identity.
    ident = threading.current_thread().ident

    # Only act on proxied threads.
    if ident not in thread_proxies:
        return

    # Read the value, close/remove the buffer, and return the value.
    thread_proxies[ident] = None
    del thread_proxies[ident]


def _get_stream(original):
    """
    Returns the inner function for use in the LocalProxy object.

    :param original: The stream to be returned if thread is not proxied.
    :type original: ``file``
    :return: The inner function for use in the LocalProxy object.
    :rtype: ``function``
    """

    def proxy():
        """
        Returns the original stream if the current thread is not proxied,
        otherwise we return the proxied item.

        :return: The stream object for the current thread.
        :rtype: ``file``
        """
        # Get the current thread's identity.
        ident = threading.current_thread().ident

        # Return the proxy, otherwise return the original.
        return thread_proxies.get(ident, original)

    # Return the inner function.
    return proxy


def enable_proxy():
    """
    Overwrites __stdout__, __stderr__, stdout, and stderr with the proxied
    objects.
    """
    sys.__stdout__ = LocalProxy(_get_stream(sys.__stdout__))
    sys.__stderr__ = LocalProxy(_get_stream(sys.__stderr__))
    sys.stdout = LocalProxy(_get_stream(sys.stdout))
    sys.stderr = LocalProxy(_get_stream(sys.stderr))


def disable_proxy():
    """
    Overwrites __stdout__, __stderr__, stdout, and stderr with the original
    objects.
    """
    sys.__stdout__ = orig___stdout__
    sys.__stderr__ = orig___stderr__
    sys.stdout = orig_stdout
    sys.stderr = orig_stderr


__all__ = [redirect, stop_redirect, enable_proxy, disable_proxy]



import unittest

from depio.Task import _get_args_dict


class TestGetArgsDict(unittest.TestCase):

    def test_get_args_dict_no_additional_args(self):
        def test_fn(a, b, c):
            pass

        result = _get_args_dict(test_fn, [1, 2, 3], {})
        self.assertEqual(result, {'a': 1, 'b': 2, 'c': 3})

    def test_get_args_dict_with_kwargs(self):
        def test_fn(a, b, c, **kwargs):
            pass

        result = _get_args_dict(test_fn, [1, 2, 3], {'d': 4, 'e': 5})
        self.assertEqual(result, {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5})

    def test_get_args_dict_with_kwargs_swapped(self):
        def test_fn(a, b, c, **kwargs):
            pass

        result = _get_args_dict(test_fn, [1, 2, 3], {'d': 5, 'e': 4})
        self.assertEqual(result, {'a': 1, 'b': 2, 'c': 3, 'd': 5, 'e': 4})

    def test_get_args_dict_with_partial_args_and_kwargs(self):
        def test_fn(a, b, c, *args, **kwargs):
            pass

        result = _get_args_dict(test_fn, [1], {'b': 2, 'c': 3, 'd': 4})
        self.assertEqual(result, {'a': 1, 'b': 2, 'c': 3, 'd': 4})


if __name__ == "__main__":
    unittest.main()



# File: test_Task__get_not_updated_products.py

from depio.Task import _get_not_updated_products


def test_get_not_updated_products_no_updates():
    product_timestamps_before_running = {'product1': 'timestamp1', 'product2': 'timestamp2'}
    product_timestamps_after_running = {'product1': 'timestamp1', 'product2': 'timestamp2'}
    not_updated_products = _get_not_updated_products(product_timestamps_after_running, product_timestamps_before_running)
    assert not_updated_products == ['product1', 'product2']


def test_get_not_updated_products_some_updates():
    product_timestamps_before_running = {'product1': 'timestamp1', 'product2': 'timestamp2'}
    product_timestamps_after_running = {'product1': 'timestamp1', 'product2': 'new_timestamp2'}
    not_updated_products = _get_not_updated_products(product_timestamps_after_running, product_timestamps_before_running)
    assert not_updated_products == ['product1']


def test_get_not_updated_products_all_updates():
    product_timestamps_before_running = {'product1': 'timestamp1', 'product2': 'timestamp2'}
    product_timestamps_after_running = {'product1': 'new_timestamp1', 'product2': 'new_timestamp2'}
    not_updated_products = _get_not_updated_products(product_timestamps_after_running, product_timestamps_before_running)
    assert not_updated_products == []


def test_get_not_updated_products_different_order():
    product_timestamps_before_running = {'product1': 'timestamp1', 'product2': 'timestamp2'}
    product_timestamps_after_running = {'product2': 'timestamp2', 'product1': 'timestamp1'}
    not_updated_products = _get_not_updated_products(product_timestamps_after_running, product_timestamps_before_running)
    assert not_updated_products == ['product1', 'product2']




import pytest
from unittest.mock import Mock
from pathlib import Path

from depio.Pipeline import Pipeline, Task, DependencyNotAvailableException

class TaskMock(Task):
    def __init__(self, func, dependencies=[], products=[]):
        self.func = func
        self.name = str(func)
        self.func_args = []
        self.func_kwargs = {}
        self.dependencies = dependencies
        self.products = products
        self.task_dependencies = []
        self.dependent_tasks = []


class PathMock(Path):
    if type(Path())._flavour:
        _flavour = type(Path())._flavour

    def __init__(self, name: str, exists: bool = True):
        # Skip the `Path`'s __init__ method
        self._exists = exists

    def exists(self) -> bool:
        return self._exists


@pytest.fixture
def pipeline():
    return Pipeline(None, False, quiet=True)

def dummyfunc(self):
    pass

# def test_pathmock():
#     p = PathMock("test")
#     assert str(p) == "test"
#     assert repr(p) == "test"
#
# def test_solve_order_no_dependency(pipeline):
#     product_A = PathMock('product_A', exists=True)
#     product_B = PathMock("product_B", exists=True)
#
#     task_A = Task("task_A", dummyfunc, produces=[product_A])
#     pipeline.add_task(task_A)
#     pipeline._solve_order()
#
#     assert task_A.task_dependencies == []
#
#
# def test_solve_order_single_dependency(pipeline):
#     product_A = PathMock('product_A', exists=True)
#     product_B = PathMock("product_B", exists=True)
#
#     task_A = TaskMock("task_A", [], [product_A])
#     task_B = TaskMock("task_B", [product_A], [product_B])
#     pipeline.add_task(task_A)
#     pipeline.add_task(task_B)
#     pipeline._solve_order()
#
#     assert task_B.task_dependencies == [task_A]
#
#
# def test_solve_order_multiple_dependencies(pipeline):
#     product_A = PathMock("product_A", exists=True)
#     product_B = PathMock("product_B", exists=True)
#     product_C = PathMock("product_C", exists=True)
#
#     task_A = TaskMock("task_A", [], [product_A])
#     task_B = TaskMock("task_B", [], [product_B])
#     task_C = TaskMock("task_C", [product_A, product_B], [product_C])
#     task_A = pipeline.add_task(task_A)
#     task_B = pipeline.add_task(task_B)
#     task_C = pipeline.add_task(task_C)
#     pipeline._solve_order()
#
#     assert sorted(task_C.task_dependencies, key=id) == sorted([task_A, task_B], key=id)
#
#
# def test_solve_order_dependency_not_available(pipeline):
#     product_A = PathMock("product_A", exists=True)
#     product_B = PathMock("product_B", exists=False)
#
#     task_A = TaskMock("task_A", [product_B], [product_A])
#
#     pipeline.add_task(task_A)
#
#     with pytest.raises(DependencyNotAvailableException):
#         pipeline._solve_order()
